@@ -1,6 +1,7 @@
 /*
  * Copyright (C) 2020 Richard Hughes <richard@hughsie.com>
  * Copyright (C) 2022 Joe Qian <joeqian@google.com>
+ * Copyright (C) 2025 Colin Kinloch <colin.kinloch@collabora.com>
  *
  * SPDX-License-Identifier: LGPL-2.1+
  */
@@ -19,6 +20,19 @@
 #include "jcat-item-private.h"
 #include "jcat-result-private.h"
 
+#ifdef HAVE_LIBCRYPTO_PKCS7
+#include "jcat-libcrypto-pkcs7-engine.h"
+#endif
+#ifdef HAVE_GNUTLS_PKCS7
+#include "jcat-gnutls-pkcs7-engine.h"
+#endif
+
+#ifdef HAVE_GNUTLS_ED25519
+#include "jcat-gnutls-ed25519-engine.h"
+#endif
+
+typedef JcatEngine *(*EngineNewFunc)(JcatContext *context);
+
 static void
 jcat_blob_func(void)
 {
@@ -102,6 +116,7 @@ jcat_file_func(void)
 	g_autoptr(JcatItem) item1 = NULL;
 	g_autoptr(JcatItem) item2 = NULL;
 	g_autoptr(JcatItem) item3 = NULL;
+	g_autoptr(JcatItem) item4 = NULL;
 	g_autoptr(JcatItem) item = jcat_item_new("firmware.bin");
 	const gchar *json_perfect = "{\n"
 				    "  \"JcatVersionMajor\" : 0,\n"
@@ -139,9 +154,9 @@ jcat_file_func(void)
 	g_assert_cmpint(jcat_blob_get_timestamp(blob2), ==, 0);
 
 	/* get default item */
-	item1 = jcat_file_get_item_default(file, &error);
+	item4 = jcat_file_get_item_default(file, &error);
 	g_assert_error(error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND);
-	g_assert_null(item1);
+	g_assert_null(item4);
 	g_clear_error(&error);
 
 	/* check item */
@@ -150,7 +165,6 @@ jcat_file_func(void)
 	g_assert_cmpint(blobs0->len, ==, 0);
 	jcat_item_add_blob(item, blob1);
 	jcat_item_add_blob(item, blob2);
-	jcat_item_add_blob(item, blob2);
 	jcat_item_add_alias_id(item, "foo.bin");
 	blobs1 = jcat_item_get_blobs(item);
 	g_assert_cmpint(blobs1->len, ==, 2);
@@ -351,7 +365,6 @@ jcat_gpg_engine_func(void)
 				  "-----END PGP SIGNATURE-----\n";
 
 	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp/libjcat-self-test/var");
 	pki_dir = g_test_build_filename(G_TEST_DIST, "pki", NULL);
 	jcat_context_add_public_keys(context, pki_dir);
 
@@ -422,7 +435,6 @@ jcat_gpg_engine_msg_func(void)
 			   "-----END PGP MESSAGE-----\n";
 
 	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp/libjcat-self-test/var");
 	pki_dir = g_test_build_filename(G_TEST_DIST, "pki", NULL);
 	jcat_context_add_public_keys(context, pki_dir);
 
@@ -450,9 +462,9 @@ jcat_gpg_engine_msg_func(void)
 }
 
 static void
-jcat_pkcs7_engine_func(void)
+jcat_pkcs7_engine_func(gconstpointer test_data)
 {
-#ifdef ENABLE_PKCS7
+#ifdef HAVE_PKCS7
 	g_autofree gchar *fn_fail = NULL;
 	g_autofree gchar *fn_pass = NULL;
 	g_autofree gchar *fn_sig = NULL;
@@ -469,12 +481,15 @@ jcat_pkcs7_engine_func(void)
 	g_autoptr(JcatResult) result_pass = NULL;
 
 	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp/libjcat-self-test/var");
 	pki_f = g_test_build_filename(G_TEST_DIST, "pki", "LVFS-CA.pem", NULL);
 	jcat_context_add_public_key(context, pki_f);
 
 	/* get engine */
-	engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_PKCS7, &error);
+	if (test_data == NULL) {
+		engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_PKCS7, &error);
+	} else {
+		engine = ((EngineNewFunc)test_data)(context);
+	}
 	g_assert_no_error(error);
 	g_assert_nonnull(engine);
 	g_assert_cmpint(jcat_engine_get_kind(engine), ==, JCAT_BLOB_KIND_PKCS7);
@@ -523,34 +538,40 @@ jcat_pkcs7_engine_func(void)
 	g_assert_null(result_fail);
 	g_clear_error(&error);
 #else
-	g_test_skip("no GnuTLS support enabled");
+	g_test_skip("no PKCS7 support enabled");
 #endif
 }
 
 static void
-jcat_pkcs7_engine_self_signed_func(void)
+jcat_pkcs7_engine_self_signed_func(gconstpointer test_data)
 {
-#ifdef ENABLE_PKCS7
-	static const char payload_str[] = "Hello, world!";
+#ifdef HAVE_PKCS7
+	static const char payload_str[] = "{\n  \"hello\": \"world\"\n}";
+	static const char payload_other_str[] = "{\n  \"hello\": \"xorld\"\n}";
+	g_autofree gchar *other_keyring_path = NULL;
 	g_autofree gchar *str = NULL;
 	g_autoptr(JcatBlob) signature = NULL;
+	g_autoptr(JcatBlob) signature_fresh = NULL;
 	g_autoptr(JcatContext) context = jcat_context_new();
 	g_autoptr(JcatEngine) engine = NULL;
 	g_autoptr(JcatEngine) engine2 = NULL;
 	g_autoptr(JcatResult) result = NULL;
+	g_autoptr(JcatResult) result_fail = NULL;
 	g_autoptr(GBytes) payload = NULL;
+	g_autoptr(GBytes) payload_other = NULL;
 	g_autoptr(GError) error = NULL;
 	const gchar *str_perfect = "JcatResult:\n"
 				   "  Timestamp:             1970-01-01T03:25:45Z\n"
-				   "  JcatPkcs7Engine:\n"
+				   "  Jcat*Pkcs7Engine:\n"
 				   "    Kind:                pkcs7\n"
 				   "    VerifyKind:          signature\n";
 
-	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp");
-
 	/* get engine */
-	engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_PKCS7, &error);
+	if (test_data == NULL) {
+		engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_PKCS7, &error);
+	} else {
+		engine = ((EngineNewFunc)test_data)(context);
+	}
 	g_assert_no_error(error);
 	g_assert_nonnull(engine);
 
@@ -575,16 +596,103 @@ jcat_pkcs7_engine_self_signed_func(void)
 	g_object_set(result, "timestamp", (gint64)12345, NULL);
 	str = jcat_result_to_string(result);
 	g_print("%s", str);
-	g_assert_cmpstr(str, ==, str_perfect);
+	g_assert_true(g_pattern_match_simple(str_perfect, str));
+
+	/* verify a payload with the wrong signature */
+	payload_other = g_bytes_new_static(payload_other_str, sizeof(payload_other_str));
+	result_fail = jcat_engine_self_verify(engine,
+					      payload_other,
+					      jcat_blob_get_data(signature),
+					      JCAT_VERIFY_FLAG_NONE,
+					      &error);
+	g_assert_error(error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA);
+	g_assert_null(result_fail);
+	g_clear_error(&error);
+
+	/* change keyring path */
+	other_keyring_path = g_dir_make_tmp("keyring_2_XXXXXX", &error);
+	g_assert_no_error(error);
+	jcat_context_set_keyring_path(context, other_keyring_path);
+
+	/* generate fresh self signing keys */
+	signature_fresh =
+	    jcat_engine_self_sign(engine, payload, JCAT_SIGN_FLAG_ADD_TIMESTAMP, &error);
+
+	/* verify original signature with original payload but new signature */
+	result_fail = jcat_engine_self_verify(engine,
+					      payload,
+					      jcat_blob_get_data(signature),
+					      JCAT_VERIFY_FLAG_NONE,
+					      &error);
+	g_assert_error(error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA);
+	g_assert_null(result_fail);
+	g_clear_error(&error);
 #else
-	g_test_skip("no GnuTLS support enabled");
+	g_test_skip("no PKCS7 support enabled");
 #endif
 }
 
+#ifdef HAVE_GNUTLS_PQC
 static void
-jcat_ed25519_engine_func(void)
+jcat_pkcs7_engine_self_signed_pq_func(gconstpointer test_data)
 {
-#ifdef ENABLE_ED25519
+	static const char payload_str[] = "Hello, world!";
+	g_autofree gchar *str = NULL;
+	g_autoptr(JcatBlob) signature = NULL;
+	g_autoptr(JcatContext) context = jcat_context_new();
+	g_autoptr(JcatEngine) engine = NULL;
+	g_autoptr(JcatEngine) engine2 = NULL;
+	g_autoptr(JcatResult) result = NULL;
+	g_autoptr(GBytes) payload = NULL;
+	g_autoptr(GError) error = NULL;
+	const gchar *str_perfect = "JcatResult:\n"
+				   "  Timestamp:             1970-01-01T03:25:45Z\n"
+				   "  JcatGnutlsPkcs7Engine:\n"
+				   "    Kind:                pkcs7\n"
+				   "    VerifyKind:          signature\n";
+
+	/* set up context */
+	jcat_context_set_keyring_path(context, "/tmp");
+
+	/* get engine */
+	if (test_data == NULL) {
+		engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_PKCS7, &error);
+	} else {
+		engine = ((EngineNewFunc)test_data)(context);
+	}
+
+	payload = g_bytes_new_static(payload_str, sizeof(payload_str));
+	g_assert_nonnull(payload);
+	signature = jcat_engine_self_sign(engine,
+					  payload,
+					  JCAT_SIGN_FLAG_ADD_TIMESTAMP | JCAT_SIGN_FLAG_USE_PQ,
+					  &error);
+	g_assert_no_error(error);
+	g_assert_nonnull(signature);
+	result = jcat_engine_self_verify(engine,
+					 payload,
+					 jcat_blob_get_data(signature),
+					 JCAT_VERIFY_FLAG_ONLY_PQ,
+					 &error);
+	g_assert_no_error(error);
+	g_assert_nonnull(result);
+
+	/* verify engine set */
+	engine2 = jcat_result_get_engine(result);
+	g_assert(engine == engine2);
+
+	/* to string */
+	g_object_set(result, "timestamp", (gint64)12345, NULL);
+	str = jcat_result_to_string(result);
+	g_print("%s", str);
+	g_assert_cmpstr(str, ==, str_perfect);
+}
+#endif
+
+static void
+jcat_ed25519_engine_func(gconstpointer test_data)
+{
+#ifdef HAVE_ED25519
 	g_autofree gchar *fn_fail = NULL;
 	g_autofree gchar *fn_pass = NULL;
 	g_autofree gchar *fn_sig = NULL;
@@ -599,12 +707,15 @@ jcat_ed25519_engine_func(void)
 	g_autoptr(JcatResult) result_pass = NULL;
 
 	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp/libjcat-self-test/var");
 	pki_dir = g_test_build_filename(G_TEST_DIST, "pki", NULL);
 	jcat_context_add_public_keys(context, pki_dir);
 
 	/* get engine */
-	engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_ED25519, &error);
+	if (test_data == NULL) {
+		engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_ED25519, &error);
+	} else {
+		engine = ((EngineNewFunc)test_data)(context);
+	}
 	g_assert_no_error(error);
 	g_assert_nonnull(engine);
 	g_assert_cmpint(jcat_engine_get_kind(engine), ==, JCAT_BLOB_KIND_ED25519);
@@ -635,33 +746,31 @@ jcat_ed25519_engine_func(void)
 	g_assert_null(result_fail);
 	g_clear_error(&error);
 #else
-	g_test_skip("no GnuTLS support enabled");
+	g_test_skip("no ED25519 support enabled");
 #endif
 }
 
 static void
-jcat_ed25519_engine_self_signed_func(void)
+jcat_ed25519_engine_self_signed_func(gconstpointer test_data)
 {
-#ifdef ENABLE_ED25519
-	static const char payload_str[] = "Hello, world!";
-	g_autofree gchar *tmp_dir = NULL;
+#ifdef HAVE_ED25519
+	static const char payload_str[] = "{\n  \"hello\": \"world\"\n}";
 	g_autoptr(JcatContext) context = jcat_context_new();
 	g_autoptr(JcatEngine) engine = NULL;
 	g_autoptr(GBytes) payload = NULL;
 	g_autoptr(GError) error = NULL;
 	const gchar *str_perfect = "JcatResult:\n"
 				   "  Timestamp:             1970-01-01T03:25:45Z\n"
-				   "  JcatEd25519Engine:\n"
+				   "  Jcat*Ed25519Engine:\n"
 				   "    Kind:                ed25519\n"
 				   "    VerifyKind:          signature\n";
 
-	/* set up context */
-	tmp_dir = g_dir_make_tmp(NULL, &error);
-	g_assert_no_error(error);
-	jcat_context_set_keyring_path(context, tmp_dir);
-
 	/* get engine */
-	engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_ED25519, &error);
+	if (test_data == NULL) {
+		engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_ED25519, &error);
+	} else {
+		engine = ((EngineNewFunc)test_data)(context);
+	}
 	g_assert_no_error(error);
 	g_assert_nonnull(engine);
 
@@ -698,17 +807,17 @@ jcat_ed25519_engine_self_signed_func(void)
 		g_object_set(result, "timestamp", (gint64)12345, NULL);
 		str = jcat_result_to_string(result);
 		g_print("%s", str);
-		g_assert_cmpstr(str, ==, str_perfect);
+		g_assert_true(g_pattern_match_simple(str_perfect, str));
 	}
 #else
-	g_test_skip("no GnuTLS support enabled");
+	g_test_skip("no ED25519 support enabled");
 #endif
 }
 
 static void
 jcat_context_verify_blob_func(void)
 {
-#ifdef ENABLE_PKCS7
+#ifdef HAVE_PKCS7
 	g_autofree gchar *fn_pass = NULL;
 	g_autofree gchar *fn_sig = NULL;
 	g_autofree gchar *pki_dir = NULL;
@@ -727,7 +836,6 @@ jcat_context_verify_blob_func(void)
 	g_autoptr(JcatResult) result_disallow = NULL;
 
 	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp");
 	pki_dir = g_test_build_filename(G_TEST_DIST, "pki", NULL);
 	jcat_context_add_public_keys(context, pki_dir);
 
@@ -780,14 +888,14 @@ jcat_context_verify_blob_func(void)
 	g_assert_error(error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED);
 	g_assert_null(result_disallow);
 #else
-	g_test_skip("no GnuTLS support enabled");
+	g_test_skip("no PKCS7 support enabled");
 #endif
 }
 
 static void
 jcat_context_verify_item_sign_func(void)
 {
-#ifdef ENABLE_PKCS7
+#ifdef HAVE_PKCS7
 	JcatResult *result;
 	g_autofree gchar *fn_pass = NULL;
 	g_autofree gchar *fn_sig = NULL;
@@ -808,7 +916,6 @@ jcat_context_verify_item_sign_func(void)
 	g_autoptr(GPtrArray) results_pass = NULL;
 
 	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp");
 	pki_dir = g_test_build_filename(G_TEST_DIST, "pki", NULL);
 	jcat_context_add_public_keys(context, pki_dir);
 
@@ -866,14 +973,14 @@ jcat_context_verify_item_sign_func(void)
 	g_assert_null(results_fail);
 	g_clear_error(&error);
 #else
-	g_test_skip("no GnuTLS support enabled");
+	g_test_skip("no PKCS7 support enabled");
 #endif
 }
 
 static void
 jcat_context_verify_item_target_func(void)
 {
-#ifdef ENABLE_PKCS7
+#ifdef HAVE_PKCS7
 	JcatResult *result;
 	g_autofree gchar *fn_pass = NULL;
 	g_autofree gchar *fn_sig = NULL;
@@ -891,7 +998,6 @@ jcat_context_verify_item_target_func(void)
 	g_autoptr(GPtrArray) results_pass = NULL;
 
 	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp");
 	pki_f = g_test_build_filename(G_TEST_BUILT, "pki", "test.pem", NULL);
 	jcat_context_add_public_key(context, pki_f);
 
@@ -936,14 +1042,14 @@ jcat_context_verify_item_target_func(void)
 	g_assert_cmpint(jcat_result_get_timestamp(result), >=, 1502871248);
 	g_assert_cmpstr(jcat_result_get_authority(result), ==, "O=Hughski Limited");
 #else
-	g_test_skip("no GnuTLS support enabled");
+	g_test_skip("no PKCS7 support enabled");
 #endif
 }
 
 static void
 jcat_context_verify_item_csum_func(void)
 {
-#ifdef ENABLE_PKCS7
+#ifdef HAVE_PKCS7
 	JcatResult *result;
 	g_autofree gchar *fn_pass = NULL;
 	g_autofree gchar *pki_dir = NULL;
@@ -964,7 +1070,6 @@ jcat_context_verify_item_csum_func(void)
 	    "a196504d09871da4f7d83b874b500f8ee6e0619ab799f074814b316d88f96f7f";
 
 	/* set up context */
-	jcat_context_set_keyring_path(context, "/tmp");
 	pki_dir = g_test_build_filename(G_TEST_DIST, "pki", NULL);
 	jcat_context_add_public_keys(context, pki_dir);
 
@@ -1016,7 +1121,7 @@ jcat_context_verify_item_csum_func(void)
 	g_assert_null(results_fail);
 	g_clear_error(&error);
 #else
-	g_test_skip("no GnuTLS support enabled");
+	g_test_skip("no PKCS7 support enabled");
 #endif
 }
 
@@ -1112,7 +1217,9 @@ jcat_bt_checkpoint_func(void)
 static void
 jcat_bt_common_func(void)
 {
+#ifdef HAVE_ED25519
 	gboolean ret;
+#endif
 	g_autofree gchar *fn_btcheckpoint = NULL;
 	g_autofree gchar *fn_btverifier = NULL;
 	g_autoptr(GBytes) blob_btcheckpoint = NULL;
@@ -1142,6 +1249,7 @@ jcat_bt_common_func(void)
 	g_assert_no_error(error);
 	g_assert_nonnull(btcheckpoint);
 
+#ifdef HAVE_ED25519
 	/* get engine */
 	engine = jcat_context_get_engine(context, JCAT_BLOB_KIND_ED25519, &error);
 	g_assert_no_error(error);
@@ -1162,6 +1270,7 @@ jcat_bt_common_func(void)
 					   &error);
 	g_assert_no_error(error);
 	g_assert_nonnull(result);
+#endif
 }
 
 int
@@ -1169,7 +1278,9 @@ main(int argc, char **argv)
 {
 	g_setenv("G_TEST_SRCDIR", SRCDIR, FALSE);
 	g_setenv("G_TEST_BUILDDIR", BUILDDIR, FALSE);
-	g_test_init(&argc, &argv, NULL);
+
+	/* ISOLATE_DIRS allows us to avoid creating keyring paths for each test */
+	g_test_init(&argc, &argv, G_TEST_OPTION_ISOLATE_DIRS, NULL);
 
 	/* only critical and error are fatal */
 	g_log_set_fatal_mask(NULL, G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL);
@@ -1186,10 +1297,43 @@ main(int argc, char **argv)
 	g_test_add_func("/jcat/engine{sha256}", jcat_sha256_engine_func);
 	g_test_add_func("/jcat/engine{gpg}", jcat_gpg_engine_func);
 	g_test_add_func("/jcat/engine{gpg-msg}", jcat_gpg_engine_msg_func);
-	g_test_add_func("/jcat/engine{pkcs7}", jcat_pkcs7_engine_func);
-	g_test_add_func("/jcat/engine{pkcs7-self-signed}", jcat_pkcs7_engine_self_signed_func);
-	g_test_add_func("/jcat/engine{ed25519}", jcat_ed25519_engine_func);
-	g_test_add_func("/jcat/engine{ed25519-self-signed}", jcat_ed25519_engine_self_signed_func);
+	g_test_add_data_func("/jcat/engine{pkcs7}", NULL, jcat_pkcs7_engine_func);
+	g_test_add_data_func("/jcat/engine{pkcs7-self-signed}",
+			     NULL,
+			     jcat_pkcs7_engine_self_signed_func);
+#ifdef HAVE_LIBCRYPTO_PKCS7
+	g_test_add_data_func("/jcat/engine{pkcs7-openssl}",
+			     &jcat_libcrypto_pkcs7_engine_new,
+			     jcat_pkcs7_engine_func);
+	g_test_add_data_func("/jcat/engine{pkcs7-self-signed-openssl}",
+			     &jcat_libcrypto_pkcs7_engine_new,
+			     jcat_pkcs7_engine_self_signed_func);
+#endif
+#ifdef HAVE_GNUTLS_PKCS7
+	g_test_add_data_func("/jcat/engine{pkcs7-gnutls}",
+			     &jcat_gnutls_pkcs7_engine_new,
+			     jcat_pkcs7_engine_func);
+	g_test_add_data_func("/jcat/engine{pkcs7-self-signed-gnutls}",
+			     &jcat_gnutls_pkcs7_engine_new,
+			     jcat_pkcs7_engine_self_signed_func);
+#endif
+#ifdef HAVE_GNUTLS_PQC
+	g_test_add_data_func("/jcat/engine{pkcs7-self-signed-pq-gnutls}",
+			     &jcat_gnutls_pkcs7_engine_new,
+			     jcat_pkcs7_engine_self_signed_pq_func);
+#endif
+	g_test_add_data_func("/jcat/engine{ed25519}", NULL, jcat_ed25519_engine_func);
+	g_test_add_data_func("/jcat/engine{ed25519-self-signed}",
+			     NULL,
+			     jcat_ed25519_engine_self_signed_func);
+#ifdef HAVE_GNUTLS_ED25519
+	g_test_add_data_func("/jcat/engine{ed25519-gnutls}",
+			     &jcat_gnutls_ed25519_engine_new,
+			     jcat_ed25519_engine_func);
+	g_test_add_data_func("/jcat/engine{ed25519-self-signed-gnutls}",
+			     &jcat_gnutls_ed25519_engine_new,
+			     jcat_ed25519_engine_self_signed_func);
+#endif
 	g_test_add_func("/jcat/context{verify-blob}", jcat_context_verify_blob_func);
 	g_test_add_func("/jcat/context{verify-item-sign}", jcat_context_verify_item_sign_func);
 	g_test_add_func("/jcat/context{verify-item-csum}", jcat_context_verify_item_csum_func);
